from pyiron_base import Project, GenericJob, GenericParameters
import numpy as np
import matplotlib.pyplot as plt
from damask import Config
from damask import Grid
from damask import Result
from damask import seeds
import h5py
import yaml
import os


class DAMASKjob(GenericJob):
    def __init__(self, project, job_name):
        super(DAMASKjob, self).__init__(project, job_name)
        self.input = GenericParameters(table_name="input")
        self.input['C11'] = 0
        self.input['C12'] = 0
        self.input['C44'] = 0
        self.input['grid'] = np.array([16,16,16])
        self.input['size'] = np.array([1.0,1.0,1.0])
        self.input['grains'] = 20
        self.executable = "DAMASK_grid -l tensionX.load -g damask.vtr"
        self._material = None 
        self._tension = None 
        
    @property
    def material(self):
        return self._material
    
    @material.setter
    def material(self, mat):
        self._material = mat
        
    @property
    def tension(self):
        return self._tension
    
    @tension.setter
    def tension(self, tension):
        self._tension = tension
        
    def write_input(self): 
        with open(os.path.join(self.working_directory, 'material.yaml'), "w") as f:
            yaml.dump(self._material, f)
        with open(os.path.join(self.working_directory, 'tensionX.load'), "w") as f:
            f.writelines(self._tension)
        seed = seeds.from_random(self.input['size'], self.input['grains'])
        new_geom = Grid.from_Voronoi_tessellation(self.input['grid'], self.input['size'], seed)
        # new_geom.save_ASCII(os.path.join(self.working_directory, "damask.geom"))
        new_geom.save(os.path.join(self.working_directory, "damask"))
        C_matrix = [self.input['C11']*1e9, self.input['C12']*1e9, self.input['C44']*1e9]
        elasticity={}
        elasticity.update({'type': 'hooke'})
        elastic_constants = {'C_11': C_matrix[0], 'C_12': C_matrix[1], 'C_44': C_matrix[2]}
        elasticity.update(elastic_constants)
        mat = Config.load(os.path.join(self.working_directory, 'material.yaml'))
        mat['phase']['Aluminum']['elasticity'] = elasticity
        mat.save(os.path.join(self.working_directory, 'material.yaml'))
        
    def collect_output(self): 
        pass 
    
    def to_hdf(self, hdf=None, group_name=None): 
        super().to_hdf(
            hdf=hdf,
            group_name=group_name
        )
        with self.project_hdf5.open("input") as h5in:
            self.input.to_hdf(h5in)
            h5in["material"] = self._material
            h5in["tension"] = self._tension

    def from_hdf(self, hdf=None, group_name=None): 
        super().from_hdf(
            hdf=hdf,
            group_name=group_name
        )
        with self.project_hdf5.open("input") as h5in:
            self.input.from_hdf(h5in)
            self._material = h5in["material"]
            self._tension = h5in["tension"]
    
    @property
    def output_file(self):
        file_name = os.path.join(self.working_directory, "damask_tensionX.hdf5")
        if self.status.finished and os.path.exists(file_name):
            return file_name
    
    @property
    def output(self):
        file_name = self.output_file
        if file_name is not None:
            return Result(file_name)

    def eval_stress(self):
        """
        return the stress as a numpy array
        Parameters
        ----------
        job_file : str
          Name of the job_file
        """
        file_name = self.output_file
        if file_name is not None:
            d = Result(file_name)
            stress_path = d.get_dataset_location('avg_sigma')
            stress = np.zeros(len(stress_path))
            hdf = h5py.File(d.fname)
            for count,path in enumerate(stress_path):
                stress[count] = np.array(hdf[path])
            stress = np.array(stress)/1E6
            return stress

    def eval_strain(self):
        """
        return the strain as a numpy array
        Parameters
        ----------
        job_file : str
          Name of the job_file
        """
        file_name = self.output_file
        if file_name is not None:
            d = Result(file_name)
            stress_path = d.get_dataset_location('avg_sigma')
            strain = np.zeros(len(stress_path))
            hdf = h5py.File(d.fname)
            for count,path in enumerate(stress_path):
                strain[count] = np.array(hdf[path.split('avg_sigma')[0]     + 'avg_epsilon'])

            return strain  
        
    def plot(self):
        """
        Plot the stress strain curve from the job file

        Parameters
        ----------
        job_file : str
        Name of the job_file
        """
        file_name = self.output_file
        if file_name is not None:
            d = Result(file_name)
            stress_path = d.get_dataset_location('avg_sigma')
            stress = np.zeros(len(stress_path))
            strain = np.zeros(len(stress_path))
            hdf = h5py.File(d.fname)
            for count, path in enumerate(stress_path):
                stress[count] = np.array(hdf[path])
                strain[count] = np.array(hdf[path.split('avg_sigma')[0]     + 'avg_epsilon'])

            stress = np.array(stress)/1E6
            plt.plot(strain,stress,linestyle='-',linewidth='2.5')
            plt.xlabel(r'$\varepsilon_{VM} $',fontsize=18)
            plt.ylabel(r'$\sigma_{VM}$ (MPa)',fontsize=18)


# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Aspect ratios of 'equal' on matplotlib 3d plots [doesn't work](https://github.com/matplotlib/matplotlib/issues/15382)
but the fenics plotting command tries to do this. I don't want to make a fork of or PR to fenics right now, so just
copy and modify the functionality
"""

import os
from dolfin.common.plotting import (
    _has_matplotlib,
    _all_plottable_types,
    _plot_x3dom,
    _matplotlib_plottable_types,
    mplot_mesh,
    mplot_dirichletbc,
    mplot_expression,
    mplot_function,
    mplot_meshfunction,
    _meshfunction_types
)
import dolfin.cpp as cpp
import ufl

__author__ = "Liam Huber"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.1"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Dec 26, 2020"


def plot(object, *args, **kwargs):
    """
    Plot given object.

    *Arguments*
        object
            a :py:class:`Mesh <dolfin.cpp.Mesh>`, a :py:class:`MeshFunction
            <dolfin.cpp.MeshFunction>`, a :py:class:`Function
            <dolfin.functions.function.Function>`, a :py:class:`Expression`
            <dolfin.cpp.Expression>, a :py:class:`DirichletBC`
            <dolfin.cpp.DirichletBC>, a :py:class:`FiniteElement
            <ufl.FiniteElement>`.

    *Examples of usage*
        In the simplest case, to plot only e.g. a mesh, simply use

        .. code-block:: python

            mesh = UnitSquare(4, 4)
            plot(mesh)

        Use the ``title`` argument to specify title of the plot

        .. code-block:: python

            plot(mesh, tite="Finite element mesh")

        It is also possible to plot an element

        .. code-block:: python

            element = FiniteElement("BDM", tetrahedron, 3)
            plot(element)

        Vector valued functions can be visualized with an alternative mode

        .. code-block:: python

            plot(u, mode = "glyphs")

        A more advanced example

        .. code-block:: python

            plot(u,
                 wireframe = True,              # use wireframe rendering
                 interactive = False,           # do not hold plot on screen
                 scalarbar = False,             # hide the color mapping bar
                 hardcopy_prefix = "myplot",    # default plotfile name
                 scale = 2.0,                   # scale the warping/glyphs
                 title = "Fancy plot",          # set your own title
                 )

    """

    # Return if plotting is disables
    if os.environ.get("DOLFIN_NOPLOT", "0") != "0":
        return

    # Return if Matplotlib is not available
    if not _has_matplotlib():
        cpp.log.info("Matplotlib is required to plot from Python.")
        return

    # Plot element
    if isinstance(object, ufl.FiniteElementBase):
        import ffc
        return ffc.plot(object, *args, **kwargs)

    # For dolfin.function.Function, extract cpp_object
    if hasattr(object, "cpp_object"):
        object = object.cpp_object()

    # Get mesh from explicit mesh kwarg, only positional arg, or via
    # object
    mesh = kwargs.pop('mesh', None)
    if isinstance(object, cpp.mesh.Mesh):
        if mesh is not None and mesh.id() != object.id():
            raise RuntimeError("Got different mesh in plot object and keyword argument")
        mesh = object
    if mesh is None:
        if isinstance(object, cpp.function.Function):
            mesh = object.function_space().mesh()
        elif hasattr(object, "mesh"):
            mesh = object.mesh()

    # Expressions do not carry their own mesh
    if isinstance(object, cpp.function.Expression) and mesh is None:
        raise RuntimeError("Expecting a mesh as keyword argument")

    backend = kwargs.pop("backend", "matplotlib")
    if backend not in ("matplotlib", "x3dom"):
        raise RuntimeError("Plotting backend %s not recognised" % backend)

    # Try to project if object is not a standard plottable type
    if not isinstance(object, _all_plottable_types):
        from dolfin.fem.projection import project
        try:
            cpp.log.info("Object cannot be plotted directly, projecting to "
                         "piecewise linears.")
            object = project(object, mesh=mesh)
            mesh = object.function_space().mesh()
            object = object._cpp_object
        except Exception as e:
            msg = "Don't know how to plot given object:\n  %s\n" \
                  "and projection failed:\n  %s" % (str(object), str(e))
            raise RuntimeError(msg)

    # Plot
    if backend == "matplotlib":
        return _plot_matplotlib(object, mesh, kwargs)
    elif backend == "x3dom":
        return _plot_x3dom(object, kwargs)
    else:
        assert False, "This code should not be reached."


def _plot_matplotlib(obj, mesh, kwargs):
    if not isinstance(obj, _matplotlib_plottable_types):
        print("Don't know how to plot type %s." % type(obj))
        return

    # Plotting is not working with all ufl cells
    if mesh.ufl_cell().cellname() not in ['interval', 'triangle', 'tetrahedron']:
        raise AttributeError(("Matplotlib plotting backend doesn't handle %s mesh.\n"
                              "Possible options are saving the output to XDMF file "
                              "or using 'x3dom' backend.") % mesh.ufl_cell().cellname())

    # Avoid importing pyplot until used
    try:
        import matplotlib.pyplot as plt
    except Exception:
        cpp.warning("matplotlib.pyplot not available, cannot plot.")
        return

    gdim = mesh.geometry().dim()
    if gdim == 3 or kwargs.get("mode") in ("warp",):
        # Importing this toolkit has side effects enabling 3d support
        from mpl_toolkits.mplot3d import axes3d  # noqa
        # Enabling the 3d toolbox requires some additional arguments
        ax = plt.gca(projection='3d')
    else:
        ax = plt.gca()
    if mesh.geometry().dim() < 3:
        ax.set_aspect('equal')

    title = kwargs.pop("title", None)
    if title is not None:
        ax.set_title(title)

    # Translate range_min/max kwargs supported by VTKPlotter
    vmin = kwargs.pop("range_min", None)
    vmax = kwargs.pop("range_max", None)
    if vmin and "vmin" not in kwargs:
        kwargs["vmin"] = vmin
    if vmax and "vmax" not in kwargs:
        kwargs["vmax"] = vmax

    # Drop unsupported kwargs and inform user
    _unsupported_kwargs = ["rescale", "wireframe"]
    for kw in _unsupported_kwargs:
        if kwargs.pop(kw, None):
            cpp.warning("Matplotlib backend does not support '%s' kwarg yet. "
                        "Ignoring it..." % kw)

    if isinstance(obj, cpp.function.Function):
        return mplot_function(ax, obj, **kwargs)
    elif isinstance(obj, cpp.function.Expression):
        return mplot_expression(ax, obj, mesh, **kwargs)
    elif isinstance(obj, cpp.mesh.Mesh):
        return mplot_mesh(ax, obj, **kwargs)
    elif isinstance(obj, cpp.fem.DirichletBC):
        return mplot_dirichletbc(ax, obj, **kwargs)
    elif isinstance(obj, _meshfunction_types):
        return mplot_meshfunction(ax, obj, **kwargs)
    else:
        raise AttributeError('Failed to plot %s' % type(obj))


# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Factories for Fenics-related object creation.
"""

import fenics as FEN
import mshr
from pyiron_base import PyironFactory

__author__ = "Liam Huber"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.1"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Dec 26, 2020"


class DomainFactory(PyironFactory):
    def __init__(self):
        super().__init__()
        self._regular = RegularMeshFactory()
        self._unit = UnitMeshFactory()

    @property
    def regular_mesh(self):
        return self._regular

    @property
    def unit_mesh(self):
        return self._unit

    def circle(self, center, radius):
        return mshr.Circle(FEN.Point(*center), radius)
    circle.__doc__ = mshr.Circle.__doc__

    def square(self, length, origin=None):
        if origin is None:
            x, y = 0, 0
        else:
            x, y = origin[0], origin[1]
        return mshr.Rectangle(FEN.Point(0 + x, 0 + y), FEN.Point(length + x, length + y))
    square.__doc__ = mshr.Rectangle.__doc__

    @staticmethod
    def box(corner1=None, corner2=None):
        """A 3d rectangular prism from `corner1` to `corner2` ((0, 0, 0) to (1, 1, 1) by default)"""
        corner1 = corner1 or (0, 0, 0)
        corner2 = corner2 or (1, 1, 1)
        return mshr.Box(FEN.Point(corner1), FEN.Point(corner2))

    @staticmethod
    def tetrahedron(p1, p2, p3, p4):
        """A tetrahedron defined by four points. (Details to be discovered and documented.)"""
        return mshr.Tetrahedron(FEN.Point(p1), FEN.Point(p2), FEN.Point(p3), FEN.Point(p4))


class UnitMeshFactory(PyironFactory):
    def square(self, nx, ny):
        return FEN.UnitSquareMesh(nx, ny)
    square.__doc__ = FEN.UnitSquareMesh.__doc__


class RegularMeshFactory(PyironFactory):
    def rectangle(self, p1, p2, nx, ny):
        return FEN.RectangleMesh(FEN.Point(p1), FEN.Point(p2), nx, ny)
    rectangle.__doc__ = FEN.RectangleMesh.__doc__

    def box(self, p1, p2, nx, ny, nz):
        return FEN.BoxMesh(FEN.Point(p1), FEN.Point(p2), nx, ny, nz)
    box.__doc__ = FEN.BoxMesh.__doc__


class BoundaryConditionFactory(PyironFactory):
    def __init__(self, job):
        self._job = job

    @staticmethod
    def _default_bc_fnc(x, on_boundary):
        return on_boundary

    def dirichlet(self, expression, bc_fnc=None):
        """
        This function defines Dirichlet boundary condition based on the given expression on the boundary.

        Args:
            expression (string): The expression used to evaluate the value of the unknown on the boundary.
            bc_fnc (fnc): The function which evaluates which nodes belong to the boundary to which the provided
                expression is applied as displacement.
        """
        bc_fnc = bc_fnc or self._default_bc_fnc
        return FEN.DirichletBC(self._job.V, expression, bc_fnc)



__version__ = "0.1"
__all__ = []


from pyiron_base import Project, JOB_CLASS_DICT

# Make classes available for new pyiron version
JOB_CLASS_DICT['Fenics'] = 'pyiron_continuum.fenics.job.generic'
JOB_CLASS_DICT['FenicsLinearElastic'] = 'pyiron_continuum.fenics.job.elastic'
JOB_CLASS_DICT['DAMASKjob'] = 'pyiron_continuum.damask.damaskjob'

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Store Fenics job plotting routines as a helper to declutter the main class.
"""

from matplotlib.docstring import copy as copy_docstring
from pyiron_continuum.fenics.fix_plotting import plot as modified_fenics_plot
import fenics as FEN
import matplotlib.pyplot as plt
import matplotlib.tri as tri
import numpy as np
from matplotlib.colors import LogNorm

__author__ = "Liam Huber"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.1"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Dec 26, 2020"


class Plot:
    """A helper class to store Fenics job plotting methods."""

    def __init__(self, job):
        """Initialize the plot helper and store the parent job."""
        self._job = job

    @copy_docstring(modified_fenics_plot)
    def plot(self, object, *args, **kwargs):
        return modified_fenics_plot(object, *args, **kwargs)

    def __call__(self, object, *args, **kwargs):
        return self.plot(object, *args, **kwargs)

    def solution(self):
        FEN.plot(self._job.solution)

    def mesh(self):
        FEN.plot(self._job.mesh)

    @staticmethod
    def _nodes_to_2d(nodes, projection_axis):
        if len(nodes) == 3:
            axes = [0, 1, 2]
            axes.pop(projection_axis)
            return nodes[axes]
        elif len(nodes) != 2:
            raise ValueError("Expected nodes to be in 2- or 3-dimensions, but got node shape {}".format(nodes.T.shape))
        else:
            return nodes

    @staticmethod
    def _nodal_values_to_1d(nodal_values):
        if len(nodal_values.shape) > 1:
            if len(nodal_values.shape) != 2:
                raise ValueError(
                    "Expected nodal values to have shape (n,) or (n, m) but got {}".format(nodal_values.shape)
                )
            nodal_values = np.linalg.norm(nodal_values, axis=-1)
        return nodal_values

    def nodal2d(
            self,
            nodal_values,
            nodes=None,
            n_grid=1000,
            n_grid_x=None,
            n_grid_y=None,
            add_colorbar=True,
            lognorm=False,
            projection_axis=None
    ):
        """
        Plot a heatmap of nodal values (or their magnitude if vectors) interpolated onto a uniform grid.

        Based off of
        [matplotlib docs](https://matplotlib.org/3.1.0/gallery/images_contours_and_fields/irregulardatagrid.html)

        Args:
            nodal_values (numpy.ndarray): A scalar array with the same length and order as the mesh points, e.g. a frame
                of scalar output. Shape should be `(n_nodes)` or `(n_nodes, n_dims)`. Note: higher dimensional values
            nodes (numpy.ndarray): xy- or xyz-positions of the nodes the values belong to. (Default is None, use mesh
                nodes.) Shape should be `(n_nodes, dim)`.
            n_grid (int): Number of points to use when interpolating the mesh values. (Default is 1000.)
            n_grid_x (int): Number of grid points to use when interpolating the mesh values in the x-direction.
                (Default is None, use n_grid value.)
            n_grid_y (int): Number of grid points to use when interpolating the mesh values in the y-direction.
                (Default is None, use n_grid value.)
            add_colorbar (bool): Whether or not to add the colorbar to the figure. (Default is True.)
            lognorm (bool): Normalize the colorscheme to a log scale. (Default is False.)
            projection_axis (int): The axis onto which to project mesh nodes and nodal values if the data is 3d.
                (Default is None, project z onto xy-plane.)

        Returns:
            (matplotlib.image.AxesImage): The imshow object.
            (matplotlib.figure.Figure): The parent figure.
            (matplotlib.axes._subplots.AxesSubplot): The subplots axis on which the plotting occurs.
        """
        n_grid_x = n_grid_x or n_grid
        n_grid_y = n_grid_y or n_grid

        nodes = self._job.mesh.coordinates() if nodes is None else nodes
        nodes = nodes.T

        projection_axis = -1 if projection_axis is None else projection_axis
        projection_axis = projection_axis % len(nodes)

        nodes = self._nodes_to_2d(nodes, projection_axis)
        nodal_values = self._nodal_values_to_1d(nodal_values)

        # Create grid values first.
        mesh_X, mesh_Y = nodes
        xi = np.linspace(np.amin(mesh_X), np.amax(mesh_X), n_grid_x)
        yi = np.linspace(np.amin(mesh_Y), np.amax(mesh_Y), n_grid_y)

        # Perform linear interpolation of the data (x,y)
        # on a grid defined by (xi,yi)
        triang = tri.Triangulation(mesh_X, mesh_Y)
        interpolator = tri.LinearTriInterpolator(triang, nodal_values)
        Xi, Yi = np.meshgrid(xi, yi)
        Zi = interpolator(Xi, Yi)

        fig, ax = plt.subplots()
        heat = ax.imshow(
            Zi[::-1],
            aspect='equal',
            cmap=plt.cm.viridis,
            extent=[Xi.min(), Xi.max(), Yi.min(), Yi.max()],
            norm=LogNorm() if lognorm else None
        )
        if add_colorbar:
            fig.colorbar(heat, shrink=0.5, aspect=10)
        return heat, fig, ax

    def nodal3d(self, nodal_values, nodes=None, add_colorbar=True):
        """
        3d scatter plot of nodal values.

        Args:
            nodal_values (numpy.ndarray): A scalar array with the same length and order as the mesh points, e.g. a frame
                of scalar output. Shape should be `(n_nodes)` or `(n_nodes, n_dims)`. Note: higher dimensional values
            nodes (numpy.ndarray): xyz-positions of the nodes the values belong to. (Default is None, use mesh nodes.)
                Shape should be `(n_nodes, 3)`.
            add_colorbar (bool): Whether or not to add the colorbar to the figure. (Default is True.)
            lognorm (bool): Normalize the colorscheme to a log scale. (Default is False.)
            projection_axis (int): The axis onto which to project mesh nodes and nodal values if the data is 3d.
                (Default is None, project z onto xy-plane.)

        Returns:
            (matplotlib.image.AxesImage): The scatter object.
            (matplotlib.figure.Figure): The parent figure.
            (matplotlib.axes._subplots.AxesSubplot): The subplots axis on which the plotting occurs.
        """
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        nodal_values = self._nodal_values_to_1d(nodal_values)
        nodes = self._job.mesh.coordinates() if nodes is None else nodes
        if nodes.shape[1] != 3:
            raise ValueError("Expected nodes to have shape (n, 3)  but got {}".format(nodes.shape))

        scattered = ax.scatter(*nodes.T, zdir='z', s=20, c=nodal_values, depthshade=True, cmap=plt.cm.viridis)
        if add_colorbar:
            fig.colorbar(scattered, shrink=0.5, aspect=10)
        return scattered, fig, ax

    def solution2d(
            self,
            frame=-1,
            n_grid=1000,
            n_grid_x=None,
            n_grid_y=None,
            add_colorbar=True,
            lognorm=False,
            projection_axis=None
    ):
        """
        Plot a heatmap of solution magnitudes interpolated onto a uniform grid.

        Args:
            frame (int): Which output frame to use. (Default is -1, most recent.)
            n_grid (int): Number of points to use when interpolating the mesh values. (Default is 1000.)
            n_grid_x (int): Number of grid points to use when interpolating the mesh values in the x-direction.
                (Default is None, use n_grid value.)
            n_grid_y (int): Number of grid points to use when interpolating the mesh values in the y-direction.
                (Default is None, use n_grid value.)
            add_colorbar (bool): Whether or not to add the colorbar to the figure. (Default is True.)
            lognorm (bool): Normalize the colorscheme to a log scale. (Default is False.)
            projection_axis (int): The axis onto which to project mesh nodes and nodal values if the data is 3d.
                (Default is None, project z onto xy-plane if needed.)

        Returns:
            (matplotlib.image.AxesImage): The imshow object.
            (matplotlib.figure.Figure): The parent figure.
            (matplotlib.axes._subplots.AxesSubplot): The subplots axis on which the plotting occurs.
        """
        return self.nodal2d(
            nodal_values=self._job.output.solution[frame],
            n_grid=n_grid,
            n_grid_x=n_grid_x,
            n_grid_y=n_grid_y,
            add_colorbar=add_colorbar,
            lognorm=lognorm,
            projection_axis=projection_axis
        )

    def solution3d(self, frame=-1, nodes=None, add_colorbar=True):
        """
        3d scatter plot of the solution magnitude.

        Args:
            frame (int): Which output frame to use. (Default is -1, most recent.)
            nodes (numpy.ndarray): xyz-positions of the nodes the values belong to. (Default is None, use mesh nodes.)
                Shape should be `(n_nodes, 3)`.
            add_colorbar (bool): Whether or not to add the colorbar to the figure. (Default is True.)

        Returns:
            (matplotlib.image.AxesImage): The scatter object.
            (matplotlib.figure.Figure): The parent figure.
            (matplotlib.axes._subplots.AxesSubplot): The subplots axis on which the plotting occurs.
        """
        return self.nodal3d(nodal_values=self._job.output.solution[frame], nodes=nodes, add_colorbar=add_colorbar)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
A job class for FEM linear elasticity with [fenics](https://fenicsproject.org/pub/tutorial/html/._ftut1008.html).
"""

from pyiron_continuum.fenics.job.generic import Fenics
from pyiron_continuum.fenics.plot import Plot

__author__ = "Liam Huber"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.1"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Dec 26, 2020"


class FenicsLinearElastic(Fenics):
    """
    Solves a linear elastic problem in three dimensions using bulk and shear moduli to describe the elasticity.
    Determines the displacement given the body load, traction, and boundary conditions.

    The variational equation solved is the integral of:
    `inner(sigma(u), epsilon(v)) * dx == dot(f, v) * dx + dot(T, v) * ds`

    Parameters:
        f (Constant/Expression): The body force term. (Default is Constant((0, 0, 0)).)
        T (Constant/Expression): The traction conditions. (Default is Constant((0, 0, 0)).)

    Input:
        bulk_modulus (float): Material elastic parameter. (Default is 76, the experimental value for Al in GPa.)
        shear_modulus (float): Material elastic parameter. (Default is 26, the experimental value for Al in GPa.)

    Output
        von_Mises (list): The von Mises stress from the mesh-evaluated solution at each step.
    """

    def __init__(self, project, job_name):
        """Create a new Fenics type job for linear elastic problems"""
        super().__init__(project=project, job_name=job_name)
        self._plot = ElasticPlot(self)

        self.input.bulk_modulus = 76
        self.input.shear_modulus = 26

        self.output.von_Mises = []

        self.V_class = self.fenics.VectorFunctionSpace

        self.f = self.Constant((0, 0, 0))
        self.T = self.Constant((0, 0, 0))

    def epsilon(self, u):
        return self.fenics.sym(self.nabla_grad(u))

    def sigma(self, u):
        lambda_ = self.input.bulk_modulus - (2 * self.input.shear_modulus / 3)
        return lambda_ * self.nabla_div(u) * self.Identity(u.geometric_dimension()) \
               + 2 * self.input.shear_modulus * self.epsilon(u)

    @property
    def LHS(self):
        return self._lhs

    @LHS.setter
    def LHS(self, _):
        raise NotImplementedError

    @property
    def RHS(self):
        return self._rhs

    @RHS.setter
    def RHS(self, _):
        raise NotImplementedError

    def validate_ready_to_run(self):
        self._lhs = self.inner(self.sigma(self.u), self.epsilon(self.v)) * self.dx
        self._rhs = self.dot(self.f, self.v) * self.dx + self.dot(self.T, self.v) * self.ds
        super().validate_ready_to_run()

    def von_Mises(self, u):
        s = self.sigma(u) - (1. / 3) * self.tr(self.sigma(u)) * self.Identity(u.geometric_dimension())
        return self.fenics.project(self.sqrt(3. / 2 * self.inner(s, s)), self.fenics.FunctionSpace(self.mesh, 'P', 1))

    def _append_to_output(self):
        super()._append_to_output()
        self.output.von_Mises.append(self.von_Mises(self.solution).compute_vertex_values(self.mesh))


class ElasticPlot(Plot):
    def stress2d(
            self,
            frame=-1,
            n_grid=1000,
            n_grid_x=None,
            n_grid_y=None,
            add_colorbar=True,
            lognorm=False,
            projection_axis=None
    ):
        """
        Plot a heatmap of the von Mises stress interpolated onto a uniform grid.

        Args:
            frame (int): Which output frame to use. (Default is -1, most recent.)
            n_grid (int): Number of points to use when interpolating the mesh values. (Default is 1000.)
            n_grid_x (int): Number of grid points to use when interpolating the mesh values in the x-direction.
                (Default is None, use n_grid value.)
            n_grid_y (int): Number of grid points to use when interpolating the mesh values in the y-direction.
                (Default is None, use n_grid value.)
            add_colorbar (bool): Whether or not to add the colorbar to the figure. (Default is True.)
            lognorm (bool): Normalize the colorscheme to a log scale. (Default is False.)
            projection_axis (int): The axis onto which to project mesh nodes and nodal values if the data is 3d.
                (Default is None, project z onto xy-plane if needed.)

        Returns:
            (matplotlib.image.AxesImage): The imshow object.
            (matplotlib.figure.Figure): The parent figure.
            (matplotlib.axes._subplots.AxesSubplot): The subplots axis on which the plotting occurs.
        """
        return self.nodal2d(
            nodal_values=self._job.output.von_Mises[frame],
            nodes=self._job.mesh.coordinates() + self._job.output.solution[frame],
            n_grid=n_grid,
            n_grid_x=n_grid_x,
            n_grid_y=n_grid_y,
            add_colorbar=add_colorbar,
            lognorm=lognorm,
            projection_axis=projection_axis
        )

    def stress3d(self, frame=-1, add_colorbar=True):
        """
        3d scatter plot of the von Mises stress.

        Args:
            frame (int): Which output frame to use. (Default is -1, most recent.)
            add_colorbar (bool): Whether or not to add the colorbar to the figure. (Default is True.)

        Returns:
            (matplotlib.image.AxesImage): The scatter object.
            (matplotlib.figure.Figure): The parent figure.
            (matplotlib.axes._subplots.AxesSubplot): The subplots axis on which the plotting occurs.
        """
        return self.nodal3d(
            nodal_values=self._job.output.von_Mises[frame],
            nodes=self._job.mesh.coordinates() + self._job.output.solution[frame],
            add_colorbar=add_colorbar
        )



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
A job class for performing finite element simulations using the [FEniCS](https://fenicsproject.org) code.
"""

import fenics as FEN
import mshr
import sympy
from pyiron_base import GenericJob, InputList
from os.path import join
import warnings
import numpy as np
from dolfin.cpp.mesh import Mesh
from pyiron_continuum.fenics.factory import DomainFactory, BoundaryConditionFactory
from pyiron_continuum.fenics.plot import Plot
from matplotlib.docstring import copy as copy_docstring
from ufl import nabla_div as ufl_nabla_div

__author__ = "Muhammad Hassani, Liam Huber"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.1"
__maintainer__ = "Muhammad Hassani"
__email__ = "hassani@mpie.de"
__status__ = "development"
__date__ = "Dec 6, 2020"


class Fenics(GenericJob):
    """
    The job provides an interface to the [FEniCS project](https://fenicsproject.org) PDE solver using the finite element
    method (FEM).

    The objective is to streamline and simplify regular usage and connect FEniCS calculations to the full job management
    and execution distribution capabilities of pyiron, without losing the power or flexibility of the underlying fenics
    library.

    Flexibility and power are currently maintained by directly exposing the underlying `fenics` and `mshr` libraries as
    attributes of the job for power users.

    Ease of use is underway, e.g. elements, trial and test functions, and the mesh are automatically populated based on
    the provided domain. Quality of life will be continuously improved as pyiron and fenics get to know each other.

    TODO: Integration with pyiron's job and data management is incomplete, as some input data types (domains and
          boundary conditions) are not yet compatible with HDF5 storage. This is a simple problem to describe, but might
          be a pain to solve with sufficient flexibility. We also need to consider storing more sophisticated output.

    TODO: Full power and flexibility still needs to be realized by allowing (a) variable function space types, (b)
          variable number of elements and trial/test functions, and (c) multiple solve types.
          (a) Is a simple input option, we just need to be smart about how to limit the choices to existing fenics
              classes.
          (b) Can probably be nicely realized by subclassing off the main job type to allow for two sets of functions --
              `V:(u,v), Q:(p,q)` -- and a variable number of functions -- `V[0]:(u[0], v[0]),...,V[n]:(u[n], v[n])` --
              which are automatically populated during mesh generation and which are accessible for building the
              equation.
          (c) Solution types just means linear system `solve(A, x, b, ...)`, linear variational problems
              `solve(a == L, u, ...)`, and nonlinear variational problems `solve(F == 0, u, ...)`. This is probably also
              going to be pretty easy to control through an input parameter with a few fixed options an a bit of
              modification to how the LHS and RHS of equations are provided, and what actually is called during `run`.
              Currently the linear variational problem is hardcoded.

    Attributes:
        input (InputList): The input parameters controlling the run.
        output (InputList): The output from the run, i.e. data that comes from `solve`ing the PDE.
        domain (?): The spatial domain on which to build the mesh or, in the case of special meshes, the mesh itself.
            To be provided prior to running the job.
        BC (?): The boundary conditions for the mesh. To be provided prior to running the job.
        LHS/RHS (?): The left-hand and right-hand sides of the equation to solve.
        time_dependent_expressions (list[Expression]): All expressions used in the domain, BC, LHS and RHS which have a
            `t` attribute that needs updating at each step. (Default is None, which initializes an empty list.)
        assigned_u (?): The term which will be assigned the solution at each timestep. (Default is None, don't assign
            anything.)
        mesh (?): The mesh. Generated automatically.
        u:
        v:
        solution:
        F:

    Input:
        mesh_resolution (int): How dense the mesh should be (larger values = denser mesh). (Default is 2, but not used
            if the domain is a special mesh, e.g. unit or regular.)
        element_type (str): What type of element should be used. (Default is 'P'.) TODO: Restrict choices.
        element_order (int): What order the elements have. (Default is 1.)  TODO: Better description.
        n_steps (int): How many steps to run for, where the `t` attribute of all time dependent expressions gets updated
            at each step. (Default is 1.)
        n_print (int): The period of steps for saving to output. (Default is 1.) Note: The final step will always be
            saved regardless of the mod value.
        dt (float): How much to increase the `t` attribute of  all time dependent expressions each step. (Default is 1.)
        solver_parameters (dict): kwargs for FEniCS solver.
            Cf. [FEniCS docs](https://fenicsproject.org/pub/tutorial/html/._ftut1017.html) for more details. (Default is
            an empty dictionary.)

    Output:
        u (list): The solved function values evaluated at the mesh points at each time step.

    Example:
        >>> job = pr.create.job.Fenics('fenics_job')
        >>> job.input.mesh_resolution = 64
        >>> job.input.element_type = 'P'
        >>> job.input.element_order = 2
        >>> job.domain = job.create.domain.circle((0, 0), 1)
        >>> job.BC = job.create.bc.dirichlet(job.Constant(0))
        >>> p = job.Expression('4*exp(-pow(beta, 2)*(pow(x[0], 2) + pow(x[1] - R0, 2)))', degree=1, beta=8, R0=0.6)
        >>> job.LHS = job.dot(job.grad_u, job.grad_v) * job.dx
        >>> job.RHS = p * job.v * job.dx
        >>> job.run()
        >>> job.plot_u()
    """

    def __init__(self, project, job_name):
        """Create a new Fenics type job"""
        super(Fenics, self).__init__(project, job_name)
        warnings.warn("Currently, the c++ dolfin functions used by fenics are not stored in the HDF5 file."
                      " This includes the domains, boundary condition, spatial functions."
                      " Therefore, it is not possible to reload the job properly, from HDF5 file."
                      " It would be safe to remove the Fenics jobs, after defining the project.")
        self._python_only_job = True
        self.create = Creator(self)
        self._plot = Plot(self)

        self.input = InputList(table_name='input')
        self.input.mesh_resolution = 2
        self.input.element_type = 'P'
        self.input.element_order = 1
        self.input.n_steps = 1
        self.input.n_print = 1
        self.input.dt = 1
        self.input.solver_parameters = {}
        # TODO?: Make input sub-classes to catch invalid input?

        self.output = InputList(table_name='output')
        self.output.solution = []

        # TODO: Figure out how to get these attributes into input/otherwise serializable
        self.domain = None  # the domain
        self.BC = None  # the boundary condition
        self._lhs = None  # the left hand side of the equation; FEniCS function
        self._rhs = None  # the right hand side of the equation; FEniCS function
        self.time_dependent_expressions = []  # Any expressions used with a `t` attribute to evolve
        # TODO: Make a class to force these to be Expressions and to update them?
        self.assigned_u = None
        self.V_class = FEN.FunctionSpace

        self._mesh = None  # the discretization mesh
        self._V = None  # finite element volume space
        self._u = None  # u is the unkown function
        self._v = None  # the test function
        self._solution = None
        self._vtk_filename = join(self.project_hdf5.path, 'output.pvd')

    # Wrap equations in setters so they can be easily protected in subclasses
    @property
    def LHS(self):
        return self._lhs

    @LHS.setter
    def LHS(self, new_lhs):
        self._lhs = new_lhs

    @property
    def RHS(self):
        return self._rhs

    @RHS.setter
    def RHS(self, new_rhs):
        self._rhs = new_rhs

    @property
    def plot(self):
        return self._plot

    def generate_mesh(self):
        if isinstance(self.domain, Mesh):
            self._mesh = self.domain  # Intent: Allow the domain to return a unit mesh
        else:
            self._mesh = mshr.generate_mesh(self.domain, self.input.mesh_resolution)

        self._V = self.V_class(self.mesh, self.input.element_type, self.input.element_order)
        # TODO: Allow changing what type of function space is used (VectorFunctionSpace, MultiMeshFunctionSpace...)
        # TODO: Allow having multiple sets of spaces and test/trial functions
        self._u = FEN.TrialFunction(self.V)
        self._v = FEN.TestFunction(self.V)
        self._solution = FEN.Function(self.V)

        if any([v is not None for v in [self.BC, self.LHS, self.RHS]]):
            warnings.warn("The mesh is being generated, but at least one of the boundary conditions or equation sides"
                          "is already defined -- please re-define these values since the mesh is updated")

    def refresh(self):
        self.generate_mesh()

    @property
    def mesh(self):
        if self._mesh is None:
            self.refresh()
        return self._mesh

    @property
    def V(self):
        if self._V is None:
            self.refresh()
        return self._V

    @property
    def u(self):
        if self._u is None:
            self.refresh()
        return self._u

    @property
    def v(self):
        if self._v is None:
            self.refresh()
        return self._v
    # TODO: Do all this refreshing with a simple decorator instead of duplicate code

    @property
    def solution(self):
        if self._solution is None:
            self.refresh()
        return self._solution

    @property
    def grad_u(self):
        return FEN.grad(self.u)

    @property
    def grad_v(self):
        return FEN.grad(self.v)

    @property
    def grad_solution(self):
        return FEN.grad(self.solution)

    @property
    def F(self):
        try:
            return self.LHS - self.RHS
        except TypeError:
            return self.LHS

    @F.setter
    def F(self, new_equation):
        self.LHS = FEN.lhs(new_equation)
        self.RHS = FEN.rhs(new_equation)

    def _write_vtk(self):
        """
        Write the output to a .vtk file.
        """
        vtkfile = FEN.File(self._vtk_filename)
        vtkfile << self.solution

    def validate_ready_to_run(self):
        if self.mesh is None:
            raise ValueError("No mesh is defined")
        if self.RHS is None:
            raise ValueError("The bilinear form (RHS) is not defined")
        if self.LHS is None:
            raise ValueError("The linear form (LHS) is not defined")
        if self.V is None:
            raise ValueError("The volume is not defined; no V defined")
        if self.BC is None:
            raise ValueError("The boundary condition(s) (BC) is not defined")

    def run_static(self):
        """
        Solve a PDE based on 'LHS=RHS' using u and v as trial and test function respectively. Here, u is the desired
        unknown and RHS is the known part.
        """
        self.status.running = True
        self._u = self.solution
        for step in np.arange(self.input.n_steps):
            for expr in self.time_dependent_expressions:
                expr.t += self.input.dt
            FEN.solve(self.LHS == self.RHS, self.u, self.BC, solver_parameters=self.input.solver_parameters)
            if step % self.input.n_print == 0 or step == self.input.n_print - 1:
                self._append_to_output()
            try:
                self.assigned_u.assign(self.solution)
            except AttributeError:
                pass
        self.status.collect = True
        self.run()

    def _append_to_output(self):
        """Evaluate the result at nodes and store in the output as a numpy array."""
        nodal_solution = self.solution.compute_vertex_values(self.mesh)
        nodes = self.mesh.coordinates()
        if len(nodal_solution) != len(nodes):
            nodal_solution = nodal_solution.reshape(nodes.T.shape).T
        self.output.solution.append(nodal_solution)

    @staticmethod
    def solution_to_original_shape(nodal_solution_frame):
        """Get a frame of the solution back to its original shape, even if it was reformed to match the mesh length."""
        if len(nodal_solution_frame.shape) == 1:
            return nodal_solution_frame
        else:
            return nodal_solution_frame.T.flatten()

    def collect_output(self):
        self._write_vtk()  # TODO: Get the output files so they're all tarballed after successful runs, like other codes
        self.to_hdf()
        self.status.finished = True

    @copy_docstring(FEN.project)
    def project_function(self, v, **kwargs):
        """
        Project v onto the job's element, V.

        Args:
            v (?): The function to project.
            **kwargs: Valid `fenics.project` kwargs (except `V`, which is provided automatically).

        Returns:
            (?): Projected function.
        """
        return FEN.project(v, V=self.V, **kwargs)

    @copy_docstring(FEN.interpolate)
    def interpolate_function(self, v):
        """
        Interpolate v on the job's element, V.

        Args:
            v (?): The function to interpolate.

        Returns:
            (?): Interpolated function.
        """
        return FEN.interpolate(v, V=self.V)

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf=hdf, group_name=group_name)
        self.input.to_hdf(hdf=self.project_hdf5)
        self.output.to_hdf(hdf=self.project_hdf5)

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(hdf=hdf, group_name=group_name)
        self.input.from_hdf(hdf=self.project_hdf5)
        self.output.from_hdf(hdf=self.project_hdf5)

    # Convenience bindings:
    @property
    def fenics(self):
        return FEN

    @property
    def mshr(self):
        return mshr

    @property
    def sympy(self):
        return sympy

    @copy_docstring(FEN.Constant)
    def Constant(self, value):
        return FEN.Constant(value)

    @copy_docstring(FEN.Expression)
    def Expression(self, *args, **kwargs):
        return FEN.Expression(*args, **kwargs)

    @copy_docstring(FEN.Identity)
    def Identity(self, dim):
        return FEN.Identity(dim)

    @property
    @copy_docstring(FEN.dx)
    def dx(self):
        return FEN.dx

    @property
    @copy_docstring(FEN.ds)
    def ds(self):
        return FEN.ds

    @copy_docstring(FEN.grad)
    def grad(self, arg):
        return FEN.grad(arg)

    @copy_docstring(FEN.nabla_grad)
    def nabla_grad(self, arg):
        return FEN.nabla_grad(arg)

    @copy_docstring(ufl_nabla_div)
    def nabla_div(self, f):
        return ufl_nabla_div(f)

    @copy_docstring(FEN.inner)
    def inner(self, a, b):
        return FEN.inner(a, b)

    @copy_docstring(FEN.dot)
    def dot(self, arg1, arg2):
        return FEN.dot(arg1, arg2)

    @copy_docstring(FEN.tr)
    def tr(self, A):
        return FEN.tr(A)

    @copy_docstring(FEN.sqrt)
    def sqrt(self, f):
        return FEN.sqrt(f)


class Creator:
    def __init__(self, job):
        self._job = job
        self._domain = DomainFactory()
        self._bc = BoundaryConditionFactory(job)

    @property
    def domain(self):
        return self._domain

    @property
    def bc(self):
        return self._bc
